# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

# cython: language_level=2

"""Module mm provides access to OS memory management interfaces like mlock and mincore."""

from posix cimport mman
from cpython.exc cimport PyErr_SetFromErrno
#from libc.stdio cimport printf

cdef extern from "<sys/user.h>":
    cpdef enum:
        PAGE_SIZE

cpdef enum:
    MLOCK_ONFAULT   = mman.MLOCK_ONFAULT
    MCL_CURRENT     = mman.MCL_CURRENT
    MCL_FUTURE      = mman.MCL_FUTURE
    MCL_ONFAULT     = mman.MCL_ONFAULT

# incore returns bytearray vector indicating whether page of mem is in core or not.
#
# mem start must be page-aligned.
def incore(const unsigned char[::1] mem not None) -> bytearray:
    cdef size_t size = mem.shape[0]
    if size == 0:
        # XXX ok? or pass addr, 0 to mincore?
        # XXX if size==0 &mem[0] raises out-of-bounds
        return bytearray()
    cdef const void *addr = &mem[0]

    # size in pages; rounded up
    cdef size_t pgsize = (size + (PAGE_SIZE-1)) // PAGE_SIZE

    #printf("\n\n%p %ld\n", addr, size)

    incore = bytearray(pgsize)
    cdef unsigned char[::1] incorev = incore

    cdef err = mman.mincore(<void *>addr, size, &incorev[0])
    if err:
        PyErr_SetFromErrno(OSError)

    return incore


# lock locks mem pages to be resident in RAM.
#
# see mlock2(2) for description of flags.
def lock(const unsigned char[::1] mem not None, int flags):
    cdef const void *addr = &mem[0]
    cdef size_t      size = mem.shape[0]

    cdef err = mman.mlock2(addr, size, flags)
    if err:
        PyErr_SetFromErrno(OSError)


# unlock unlocks mem pages from being pinned in RAM.
def unlock(const unsigned char[::1] mem not None):
    cdef const void *addr = &mem[0]
    cdef size_t      size = mem.shape[0]

    cdef err = mman.munlock(addr, size)
    if err:
        PyErr_SetFromErrno(OSError)


from posix.types cimport off_t

# XXX PROT_READ | MAP_SHARED
def map_ro(int fd, off_t offset, size_t size):
    cdef void *addr

    addr = mman.mmap(NULL, size, mman.PROT_READ, mman.MAP_SHARED, fd, offset)
    if addr == mman.MAP_FAILED:
        PyErr_SetFromErrno(OSError)

    return <unsigned char[:size:1]>addr


# unmap unmaps memory coverd by mem.
def unmap(const unsigned char[::1] mem not None):
    cdef const void *addr = &mem[0]
    cdef size_t      size = mem.shape[0]

    cdef err = mman.munmap(<void *>addr, size)
    if err:
        PyErr_SetFromErrno(OSError)
    # ok
